import logging
import os
import tarfile
import tempfile

import magic
import zstandard
from arkindex.ponos.models import Task

logger = logging.getLogger(__name__)


def is_admin_or_ponos_task(request):
    return request.user.is_authenticated and (request.user.is_admin or isinstance(request.auth, Task))


def get_process_from_task_auth(request):
    if isinstance(request.auth, Task):
        return request.auth.process


def upload_artifact(task, path, artifacts_dir):
    content_type = magic.from_file(path, mime=True)
    size = os.path.getsize(path)
    artifact = task.artifacts.create(
        path=os.path.relpath(path, artifacts_dir),
        content_type=content_type,
        size=size,
    )
    artifact.s3_object.upload_file(str(path), ExtraArgs={"ContentType": content_type})


def decompress_zst_archive(compressed_archive):
    """Decompress a zst-compressed tar archive in data dir.
    This returns the path to the archive and the file descriptor.

    .. warning::

        Beware of closing the file descriptor explicitly or the main
        process will keep the memory held even if the file is deleted.

    :param compressed_archive: Path to the target ZST-compressed archive
    :type compressed_archive: str
    :return: File descriptor and path to the uncompressed tar archive
    :rtype: list
    """
    dctx = zstandard.ZstdDecompressor()
    archive_fd, archive_path = tempfile.mkstemp(prefix="ponos-", suffix=".tar")

    logger.debug(f"Uncompressing downloaded File to {archive_path}")
    try:
        with open(compressed_archive, "rb") as compressed, open(
            archive_path, "wb"
        ) as decompressed:
            dctx.copy_stream(compressed, decompressed)
        logger.debug(f"Successfully uncompressed archive {compressed_archive}")
    except zstandard.ZstdError as e:
        raise Exception(f"Couldn't uncompressed archive: {e}")

    # Deleting ZSTD archive tempfile as soon as possible
    try:
        os.remove(compressed_archive)
        logger.debug(
            f"Successfully deleted compressed archive file {compressed_archive}"
        )
    except OSError as e:
        logger.debug(
            f"Unable to delete compressed archive file {compressed_archive}: {e}"
        )
    return archive_fd, archive_path


def extract_tar_archive(archive_fd, archive_path, destination):
    """Extract the tar archive's content to a specific destination

    :param archive_fd: File descriptor of the archive
    :type archive_fd: int
    :param archive_path: Path to the archive
    :type archive_path: str
    :param destination: Path where the archive's data will be extracted
    :type destination: str
    """
    try:
        with tarfile.open(archive_path) as tar_archive:
            tar_archive.extractall(destination)
    except tarfile.ReadError as e:
        raise Exception(f"Couldn't handle the decompressed Tar archive: {e}")
    finally:
        # Deleting Tar archive
        try:
            os.close(archive_fd)
            os.remove(archive_path)
        except OSError as e:
            logger.warning(f"Unable to delete archive file {archive_path}: {e}")
