from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import (
    ProcessMode,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase


@override_settings(PONOS_MAXIMUM_TASK_TTL=3600)
class TestWorkerRunTTL(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=Farm.objects.first(),
        )
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")
        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_run = cls.process.worker_runs.create(version=cls.dla, ttl=0)

    def test_create_default_ttl(self):
        self.client.force_login(self.superuser)
        # Corpus TTL / WorkerRun TTL
        cases = [
            (0, 0),
            (10000, 10000),
            # No corpus TTL means the instance wide value should be set
            (None, 3600),
        ]

        for corpus_ttl, expected_ttl in cases:
            with self.subTest(corpus_ttl=corpus_ttl):
                self.process.worker_runs.filter(version=self.recognizer).delete()
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process.id)}),
                        {"worker_version_id": str(self.recognizer.id)},
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                self.assertEqual(data["ttl"], expected_ttl)
                run = self.process.worker_runs.get(id=data["id"])
                self.assertEqual(run.ttl, expected_ttl)

    def test_create_set_ttl(self):
        self.client.force_login(self.superuser)
        # Corpus TTL / WorkerRun TTL / Expected WorkerRun TTL
        cases = [
            (0, 0, 0),
            (0, 1000, 1000),
            (1800, 1000, 1000),
            (1800, 1800, 1800),
            # No corpus TTL means the instance wide value is the limit
            (None, 600, 600),
            (None, 3600, 3600),
        ]

        for corpus_ttl, worker_run_ttl, expected_ttl in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.process.worker_runs.filter(version=self.recognizer).delete()
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process.id)}),
                        {
                            "worker_version_id": str(self.recognizer.id),
                            "ttl": worker_run_ttl,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                self.assertEqual(data["ttl"], expected_ttl)
                run = self.process.worker_runs.get(id=data["id"])
                self.assertEqual(run.ttl, expected_ttl)

    def test_create_invalid_ttl(self):
        self.client.force_login(self.superuser)
        self.process.worker_runs.filter(version=self.recognizer).delete()

        # Corpus TTL, WorkerRun TTL, error message
        cases = [
            (None, "one hour", ["A valid integer is required."]),
            (None, -1, ["Ensure this value is greater than or equal to 1."]),
            (None, 0, ["Ensure this value is greater than or equal to 1."]),
            (None, 1e12, ["Ensure this value is less than or equal to 3600."]),
            (0, -1, ["Ensure this value is greater than or equal to 0."]),
            (0, 1e12, ["Ensure this value is less than or equal to 2147483647."]),
            (1800, -1, ["Ensure this value is greater than or equal to 1."]),
            (1800, 0, ["Ensure this value is greater than or equal to 1."]),
            (1800, 1e12, ["Ensure this value is less than or equal to 1800."]),
        ]
        for corpus_ttl, worker_run_ttl, expected_error in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(4):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process.id)}),
                        {
                            "worker_version_id": str(self.recognizer.id),
                            "ttl": worker_run_ttl,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "ttl": expected_error,
                })

    def test_partial_update_set_ttl(self):
        self.client.force_login(self.superuser)
        # Corpus TTL / WorkerRun TTL / Expected WorkerRun TTL
        cases = [
            (0, 0, 0),
            (0, 1000, 1000),
            (1800, 1000, 1000),
            (1800, 1800, 1800),
            # No corpus TTL means the instance wide value is the limit
            (None, 600, 600),
            (None, 3600, 3600),
        ]

        for corpus_ttl, worker_run_ttl, expected_ttl in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(5):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.worker_run.id)}),
                        {"ttl": worker_run_ttl},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                self.assertEqual(data["ttl"], expected_ttl)
                run = self.process.worker_runs.get(id=data["id"])
                self.assertEqual(run.ttl, expected_ttl)

    def test_partial_update_invalid_ttl(self):
        self.client.force_login(self.superuser)

        # Corpus TTL, WorkerRun TTL, error message
        cases = [
            (None, "one hour", ["A valid integer is required."]),
            (None, -1, ["Ensure this value is greater than or equal to 1."]),
            (None, 0, ["Ensure this value is greater than or equal to 1."]),
            (None, 1e12, ["Ensure this value is less than or equal to 3600."]),
            (0, -1, ["Ensure this value is greater than or equal to 0."]),
            (0, 1e12, ["Ensure this value is less than or equal to 2147483647."]),
            (1800, -1, ["Ensure this value is greater than or equal to 1."]),
            (1800, 0, ["Ensure this value is greater than or equal to 1."]),
            (1800, 1e12, ["Ensure this value is less than or equal to 1800."]),
        ]
        for corpus_ttl, worker_run_ttl, expected_error in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(3):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.worker_run.id)}),
                        {"ttl": worker_run_ttl},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "ttl": expected_error,
                })


    def test_update_default_ttl(self):
        self.client.force_login(self.superuser)
        # Corpus TTL / WorkerRun TTL
        cases = [
            (0, 0),
            (10000, 10000),
            # No corpus TTL means the instance wide value should be set
            (None, 3600),
        ]

        for corpus_ttl, expected_ttl in cases:
            with self.subTest(corpus_ttl=corpus_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(5):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.worker_run.id)}),
                        {},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json()["ttl"], expected_ttl)
                self.worker_run.refresh_from_db()
                self.assertEqual(self.worker_run.ttl, expected_ttl)

    def test_update_set_ttl(self):
        self.client.force_login(self.superuser)
        # Corpus TTL / WorkerRun TTL / Expected WorkerRun TTL
        cases = [
            (0, 0, 0),
            (0, 1000, 1000),
            (1800, 1000, 1000),
            (1800, 1800, 1800),
            # No corpus TTL means the instance wide value is the limit
            (None, 600, 600),
            (None, 3600, 3600),
        ]

        for corpus_ttl, worker_run_ttl, expected_ttl in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(5):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.worker_run.id)}),
                        {"ttl": worker_run_ttl},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                self.assertEqual(data["ttl"], expected_ttl)
                run = self.process.worker_runs.get(id=data["id"])
                self.assertEqual(run.ttl, expected_ttl)

    def test_update_invalid_ttl(self):
        self.client.force_login(self.superuser)

        # Corpus TTL, WorkerRun TTL, error message
        cases = [
            (None, "one hour", ["A valid integer is required."]),
            (None, -1, ["Ensure this value is greater than or equal to 1."]),
            (None, 0, ["Ensure this value is greater than or equal to 1."]),
            (None, 1e12, ["Ensure this value is less than or equal to 3600."]),
            (0, -1, ["Ensure this value is greater than or equal to 0."]),
            (0, 1e12, ["Ensure this value is less than or equal to 2147483647."]),
            (1800, -1, ["Ensure this value is greater than or equal to 1."]),
            (1800, 0, ["Ensure this value is greater than or equal to 1."]),
            (1800, 1e12, ["Ensure this value is less than or equal to 1800."]),
        ]
        for corpus_ttl, worker_run_ttl, expected_error in cases:
            with self.subTest(corpus_ttl=corpus_ttl, worker_run_ttl=worker_run_ttl):
                self.corpus.maximum_task_ttl = corpus_ttl
                self.corpus.save()

                with self.assertNumQueries(3):
                    response = self.client.put(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.worker_run.id)}),
                        {"ttl": worker_run_ttl},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "ttl": expected_error,
                })
