from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Worker, WorkerType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerPartialUpdate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_type_dla = WorkerType.objects.get(slug="dla")
        cls.worker_reco = Worker.objects.get(slug="reco")

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_requires_admin(self, max_level_mock, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Contributor.value})
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
        ]

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this worker."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])
        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.worker_reco))

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update(self, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all()
        ]

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                    "type": "dla",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "Recognizer",
            "slug": "reco",
            "description": "New description",
            "type": "dla",
            "repository_url": None,
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "Recognizer")
        self.assertEqual(self.worker_reco.slug, "reco")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_unique_slug(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "slug": "dla",
                },
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "non_field_errors": ["A worker with the same repository URL and slug already exists."],
        })

    def test_archived(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.assertTrue(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json(), {
                    "id": str(self.worker_reco.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "description": "",
                    "type": "recognizer",
                    "repository_url": None,
                    "archived": new_value,
                })

    def test_new_type(self):
        """
        A new type is automatically created if necessary
        """
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {"type": "new_type"},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            "archived": False,
            "description": "",
            "id": str(self.worker_reco.id),
            "name": "Recognizer",
            "repository_url": None,
            "slug": "reco",
            "type": "new_type",
        })
        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.type.slug, "new_type")
        self.assertEqual(self.worker_reco.type.display_name, "New_type")
