from datetime import datetime
from unittest.mock import MagicMock, call, patch
from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rq.job import JobStatus

from arkindex.project.tests import FixtureAPITestCase


class MockedJob:

    def __init__(self, id=None, user_id=None, status=JobStatus.QUEUED, **kwargs):
        self.id = id or str(uuid4())
        self.enqueued_at = datetime(2020, 1, 1, 13, 37, 42)
        self.started_at = None
        self.ended_at = None
        self.progress = None
        self.description = "something"
        self.user_id = user_id
        self._status = status

        self.delete = MagicMock()
        self.restore = MagicMock()

    def get_status(self, refresh=True):
        # Those endpoints should not be setting refresh to True as they do not have to reload from Redis
        assert refresh is False, "refresh should be set to False"
        return self._status

    @classmethod
    def key_for(cls, id):
        return f"rq:job:{id}".encode()


class TestJobs(FixtureAPITestCase):

    def setUp(self):
        super().setUp()
        self.user_mocked_job = MockedJob(user_id=self.user.id)
        self.superuser_mocked_job = MockedJob(user_id=self.superuser.id)

    @patch("arkindex.users.models.get_queue")
    def test_list_requires_login(self, get_queue_mock):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:jobs-list"))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(get_queue_mock.called)

    @patch("arkindex.users.models.get_queue")
    def test_list_requires_verified(self, get_queue_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-list"))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(get_queue_mock.called)

    @override_settings(REDIS_ZREM_CHUNK_SIZE=2)
    @patch("arkindex.users.models.get_queue")
    def test_list(self, get_queue_mock):
        get_queue_mock().user_registry().key = "user_registry_key"
        # Job IDs for the current user
        get_queue_mock().user_registry().get_job_ids.return_value = ["job1", "job2", "job3", "job4", "job5"]
        # Result of all the HGETALL commands: payloads of each job
        # Three jobs do not exist, which should result in two ZREM calls
        pipeline = get_queue_mock().connection.pipeline().__enter__()
        pipeline.execute.return_value = [
            None,
            {"id": "job2", "foo": "bar"},
            None,
            {"id": "job4", "foo": "bar"},
            None,
        ]
        get_queue_mock().job_class = MockedJob
        # Reset call counts preserving the mocked return values
        get_queue_mock.reset_mock()

        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json(), [
            {
                "id": "job2",
                "status": "queued",
                "enqueued_at": "2020-01-01T13:37:42Z",
                "started_at": None,
                "ended_at": None,
                "progress": None,
                "description": "something"
            },
            {
                "id": "job4",
                "status": "queued",
                "enqueued_at": "2020-01-01T13:37:42Z",
                "started_at": None,
                "ended_at": None,
                "progress": None,
                "description": "something"
            }
        ])

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock.call_args, call("default"))
        self.assertEqual(get_queue_mock().user_registry.call_count, 1)
        self.assertEqual(get_queue_mock().user_registry.call_args, call(self.user.id))
        self.assertEqual(get_queue_mock().user_registry().get_job_ids.call_count, 1)
        self.assertEqual(get_queue_mock().user_registry().get_job_ids.call_args, call())

        self.assertListEqual(pipeline.hgetall.call_args_list, [
            call(b"rq:job:job1"),
            call(b"rq:job:job2"),
            call(b"rq:job:job3"),
            call(b"rq:job:job4"),
            call(b"rq:job:job5"),
        ])
        self.assertEqual(pipeline.execute.call_count, 1)
        self.assertEqual(pipeline.execute.call_args, call())

        self.assertEqual(get_queue_mock().connection.zrem.call_count, 2)
        chunk_call, remaining_call = get_queue_mock().connection.zrem.call_args_list
        # The chunk call is made using a set(), because we do not want to send duplicate keys,
        # so the arguments might be either job1,job3 or job3,job1
        self.assertIn(chunk_call, (
            call("user_registry_key", "job1", "job3"),
            call("user_registry_key", "job3", "job1"),
        ))
        self.assertEqual(remaining_call, call("user_registry_key", "job5"))

    @patch("arkindex.users.api.get_queue")
    def test_retrieve_requires_login(self, get_queue_mock):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(get_queue_mock.called)

    @patch("arkindex.users.api.get_queue")
    def test_retrieve_requires_verified(self, get_queue_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(get_queue_mock.called)

    @patch("arkindex.users.api.get_queue")
    def test_retrieve(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = self.user_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": self.user_mocked_job.id,
            "status": "queued",
            "enqueued_at": "2020-01-01T13:37:42Z",
            "started_at": None,
            "ended_at": None,
            "progress": None,
            "description": "something"
        })

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock.call_args, call("default"))
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call(self.user_mocked_job.id))

    @patch("arkindex.users.api.get_queue")
    def test_retrieve_custom_id(self, get_queue_mock):
        """
        Job IDs can be any string, including one with slashes, so we should allow any string in RetrieveJob
        """
        job = MockedJob(user_id=self.user.id, id="a/b/c/d")
        get_queue_mock.return_value.fetch_job.return_value = job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": "a/b/c/d"}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": "a/b/c/d",
            "status": "queued",
            "enqueued_at": "2020-01-01T13:37:42Z",
            "started_at": None,
            "ended_at": None,
            "progress": None,
            "description": "something"
        })

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock.call_args, call("default"))
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call("a/b/c/d"))

    @patch("arkindex.users.api.get_queue")
    def test_retrieve_not_found(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = None
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:jobs-retrieve", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(get_queue_mock.call_args_list, [
            call("default"),
            call("high"),
            call("export"),
        ])

        # Called once per queue
        self.assertListEqual(get_queue_mock().fetch_job.call_args_list, [
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa")
        ])

    @patch("arkindex.users.api.get_queue")
    def test_retrieve_wrong_user(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = self.superuser_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": self.superuser_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call(self.superuser_mocked_job.id))

    @patch("arkindex.users.api.get_queue")
    def test_destroy_requires_login(self, get_queue_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(get_queue_mock.called)

    @patch("arkindex.users.api.get_queue")
    def test_destroy_requires_verified(self, get_queue_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(get_queue_mock.called)

    @patch("arkindex.users.api.get_queue")
    def test_destroy(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = self.user_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:jobs-retrieve", kwargs={"pk": self.user_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call(self.user_mocked_job.id))
        self.assertEqual(self.user_mocked_job.delete.call_count, 1)

    @patch("arkindex.users.api.get_queue")
    def test_destroy_not_found(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = None
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse("api:jobs-retrieve", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(get_queue_mock.call_args_list, [
            call("default"),
            call("high"),
            call("export"),
        ])

        self.assertListEqual(get_queue_mock().fetch_job.call_args_list, [
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            call("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa")
        ])

    @patch("arkindex.users.api.get_queue")
    def test_destroy_wrong_user(self, get_queue_mock):
        get_queue_mock.return_value.fetch_job.return_value = self.superuser_mocked_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:jobs-retrieve", kwargs={"pk": self.superuser_mocked_job.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call(self.superuser_mocked_job.id))
        self.assertFalse(self.superuser_mocked_job.delete.called)

    @patch("arkindex.users.api.get_queue")
    def test_destroy_not_started(self, get_queue_mock):
        started_job = MockedJob(user_id=self.user.id, status=JobStatus.STARTED)
        get_queue_mock.return_value.fetch_job.return_value = started_job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:jobs-retrieve", kwargs={"pk": started_job.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertListEqual(response.json(), ["Cannot delete a running job."])

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call(started_job.id))
        self.assertFalse(started_job.delete.called)

    @patch("arkindex.users.api.get_queue")
    def test_destroy_custom_id(self, get_queue_mock):
        """
        Job IDs can be any string, including one with slashes, so we should allow any string in DestroyJob
        """
        job = MockedJob(user_id=self.user.id, id="a/b/c/d")
        get_queue_mock.return_value.fetch_job.return_value = job
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:jobs-retrieve", kwargs={"pk": "a/b/c/d"}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(get_queue_mock.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_count, 1)
        self.assertEqual(get_queue_mock().fetch_job.call_args, call("a/b/c/d"))
        self.assertEqual(job.delete.call_count, 1)
